\RAY.XPL	JAN-08-90
\A Simple Ray Tracer
\ by Loren Blaney
\
\This is based on an article in Byte Magazine, December 1990, page 263.
\
\Vectors are the key to understanding this program.
\
\A right handed, 3-D coordinate system is used with the origin (O) at the
\ upper-left corner of the display screen.
\
\				  Z (into screen)
\			        /
\			      /
\			    /
\			  /
\		       O+------------------- X
\			|
\			|
\			|
\			|		V
\			|	      /
\			|	    /
\			Y	  /
\			     P0 *
\
\The viewer is centered in front of the screen and back some distance, which
\ is negative Z. Objects are behind the screen with positive Z coordinates.
\
\A light ray is represented by a starting position (P0x, P0y, P0z), and a
\ direction vector (Vx, Vy, Vz). Rays are traced backwards, from the eye to
\ the objects. This makes the program much faster by eliminating rays that
\ the viewer never sees. This works because the path that a ray traces is
\ the same whether it is coming or going.

code	CHOUT=8,	KEYHIT=33,	SOFTINT=34,	GETREG=35,
	POINT=41,	LINE=42,	MOVE=43,	SETVID=45,
	FIX=50;
code real
	RLRES=46,	FLOAT=49,	RLABS=51,	SQRT=53;

def	TVW=640, TVH=480; \Screen width and height (also see SETVID in MAIN)
def	SIZE0=32;	\Initial size of "blocks" (must be power of 2 [1..32])
def	BIG=1E20, SMALL=0.01;		\Relative to size of our objects
def	BEL=$07, ESC=$1B, CTRL=$40;	\Ascii characters
def \COLOR\ BLACK, BLUE, GREEN, CYAN, RED, MAGENTA, BROWN, WHITE,
	GRAY, LBLUE, LGREEN, LCYAN, LRED, LMAGENTA, YELLOW, LWHITE, MIRROR;
int	CPUREG,		\Pointer to CPU registers. For calls to BIOS routines
	MODE0,		\Video mode when this program started
	OBJECTS,	\Number of objects (spheres)
	SPCOLOR,	\Array: color of each sphere
	REDRAW;		\Flag: redraw the image (because the view has changed)
real	SPCENTER,	\2-D array: center coordinates for each sphere
	SPRADIUS,	\Array: radius of each sphere
	EYE,		\Array: coordinates of viewer
	GAIN,		\Magnification of screen image
	XOFFSET,	\Screen image position (horizontal and vertical)
	YOFFSET;

\----------------------- FUNDAMENTAL ROUTINES -------------------------

func	GETKEY;		\Get character from keyboard.
int	CH;		\This is a low-level routine with no echo,
begin			\ no ctrl-C, and no cursor.
CPUREG(0):= 0;		\Requires: CPUREG:= GETREG;   Function = $00
SOFTINT($16);		\Call BIOS routine
CH:= CPUREG(0) & $FF;
if CH = 0 then
	begin			\Handle special keys
	case CPUREG(0)>>8 of	\Arrows:
	  72:	CH:= $0B;	\ up
	  80:	CH:= $0A;	\ down
	  75:	CH:= $08;	\ left
	  77:	CH:= $15	\ right
	other	CH:= -(CPUREG(0)>>8);
	end;
return CH;
end;	\GETKEY



func	GETVID;		\Returns the current video mode
begin			\Requires: CPUREG:= GETREG;
CPUREG(0):= $0F00;	\Function $0F
SOFTINT($10);		\Call BIOS routine
return CPUREG(0) & $FF;
end;	\GETVID

\----------------------------------------------------------------------

proc	DOKEY;		\Handle keyboard commands
int	CH;
begin
CH:= GETKEY;
REDRAW:= true;		\Anticipate that a redraw will be necessary
case CH of
  ^.:	GAIN:= GAIN *1.1;			\Zoom in a little
  ^,:	GAIN:= GAIN /1.1;			\Zoom out a little
  ^>:	GAIN:= GAIN *2.0;			\Zoom in to double size
  ^<:	GAIN:= GAIN /2.0;			\Zoom out to half size
  $15:	XOFFSET:= XOFFSET -32.0;		\Shift to the left
  $08:	XOFFSET:= XOFFSET +32.0;		\Shift to the right
  $0A:	YOFFSET:= YOFFSET -32.0;		\Shift up
  $0B:	YOFFSET:= YOFFSET +32.0;		\Shift down (Byte, byte, byte!)
  ^C-CTRL, ESC:  [SETVID(MODE0);   exit];	\Restore video mode and exit
  ^S-CTRL: [CH:= GETKEY;   REDRAW:=false]	\Stop/start program
other	[CHOUT(0, BEL);   REDRAW:= false];	\Illegal command -- beep
end;	\DOKEY



proc	BLOCK(X0, Y0, COLOR, SIZE);	\Plot a square block
int	X0, Y0,	\Coordinates of upper-left corner of block
	COLOR,	\Color of the block
	SIZE;	\Number of pixels on a side (of the block)
int	X, Y1;
begin
if SIZE = 1 then POINT(X0, Y0, COLOR)		\(For speed)
else	begin
	Y1:= Y0 +SIZE -1;				\ Y0 ---|||--
	for X:= X0, X0+SIZE-1 do			\       |||
		[MOVE(X, Y0);   LINE(X, Y1, COLOR)];	\       |||
	end;						\ Y1 ---VVV--
end;	\BLOCK

\------------------------- VECTOR ROUTINES ----------------------------

func real DOTPROD(A, B);	\Return the dot product of vectors A and B
real	A, B;			\A.B = |A| *|B| *COS(ANGLE)
begin				\    = Ax*Bx + Ay*By + Az*Bz
return A(0)*B(0) + A(1)*B(1) + A(2)*B(2);
end;	\DOTPROD



proc	UNITIZE(V);	\Convert V to a unit vector (i.e. length = 1.0)
real	V;
real	L;
begin
L:= SQRT(V(0)*V(0) + V(1)*V(1) + V(2)*V(2));	\Length of V
V(0):= V(0) /L;
V(1):= V(1) /L;
V(2):= V(2) /L;
end;	\UNITIZE



func real INTPLANE(RP, RV, Y0);	\Intersection of a ray and the X-Z plane.
\Returns the distance from RP to the point of intersection. The coordinates
\ of this point of intersection can be obtained by: P = RP + RV *DIST.
\Returns 0.0 if there is no intersection.
real	RP,	\Array: Ray point
	RV,	\Array: Ray vector
	Y0;	\Y position of X-Z plane
real	DIST;	\Distance along RV from RP to point of intersection
begin
\DIST = (Y0 - P0y) /Vy
if RLABS(RV(1)) < SMALL then return 0.0;	\Avoid divide by 0.0 error
DIST:= (Y0 - RP(1)) /RV(1);

if DIST > 0.0 then return DIST;	\The usual case
return 0.0;			\The plane is behind the ray
end;	\INTPLANE



func real INTSPHERE(RP, RV, S, R);	\Intersection of a ray and a sphere.
\Returns the distance from RP to the point of intersection. The coordinates
\ of this point of intersection can be obtained by: P = RP + RV *DIST.
\Returns 0.0 if there is no intersection.
real	RP,	\Array: Ray point
	RV,	\Array: Ray vector
	S,	\Array: Sphere center
	R;	\Radius of sphere
real	A, B, C, D, G,		\Intermediate variables
	DIST, DIST1, DIST2;	\Distances to points of intersection
begin
G:= RLRES(3);

A:= DOTPROD(RV, RV);

G(0):= RP(0) -S(0);			\G:= P0 - S;
G(1):= RP(1) -S(1);
G(2):= RP(2) -S(2);

B:= DOTPROD(RV, G);   B:= B +B;		\B:= 2 *(RV.G)

C:= DOTPROD(G, G) - R *R;

D:= B *B - 4.0 *A *C;
if D <= 0.0 then return 0.0;		\Ray does not intersect sphere

D:= SQRT(D);
DIST1:= (-B +D) /(A +A);
DIST2:= (-B -D) /(A +A);

DIST:= if DIST1 < DIST2 then DIST1 else DIST2;	\Get the closer point
if DIST > 0.0 then return DIST;		\The usual case
if DIST1 > 0.0 then return DIST1;	\The ray starts inside the sphere
if DIST2 > 0.0 then return DIST2;
return 0.0;				\The sphere is behind the ray
end;	\INTSPHERE



proc	REFLECT(I, N, R);	\Reflect a ray
real	I, N, R;	\Unit vectors: Incident, Normal, Reflected
real	D;
begin
D:= DOTPROD(I, N);			\R = I - 2 *(I.N) *N
D:= D +D;

R(0):= I(0) - N(0) *D;
R(1):= I(1) - N(1) *D;
R(2):= I(2) - N(2) *D;
end;	\REFLECT

\----------------------------- THE HEART ------------------------------

proc	TRACERAYS;	\Determine the correct color for each pixel and plot it
real	RAYV;		\Ray vector
int	X, Y,		\Pixel coordinates on the screen
	SIZE,		\Block size (resolution control)
	COLOR,		\Pixel color
	LEVEL,		\Level of recursion
	XF, YF,		\Flags used to plot smaller and smaller blocks
	FIRSTTIME;


proc	TRACE(RP, RV);	\Trace ray from RP in direction RV.
\Set COLOR to the color of the closest object intersected by this ray.
\COLOR is left unchanged if no object is intersected.
real	RP, RV;		\Ray point, Ray vector
real	RP2, RV2,	\Ray point, Ray vector for reflected ray, if any
	NORM,		\Vector normal (perpendicular) to sphere at point
			\ of intersection
	DIST, DISTMIN;	\Distance to point of intersection, Closest distance
int	OBJ, OBJMIN;	\Object, Closest object


	proc	CHECKER(D);	\Make checker-board pattern for X-Z plane
	\Inputs RP and RV
	real	D;	\Distance along RV from RP to point of intersection
	real	PX, PZ;
	int	IPX, IPZ;
	def	CHECKSIZE=1000.0;	\Size of checker squares
	begin
	\Px = P0x + Vx *D		Find point of intersection
	PX:= RP(0) + RV(0) *D;		PZ:= RP(2) + RV(2) *D;
	IPX:= FIX(PX /CHECKSIZE);	IPZ:= FIX(PZ /CHECKSIZE);
	COLOR:= if (IPX&$03)= $03 ! (IPZ&$03)=$03 then BLUE else WHITE;
	end;	\CHECKER


begin	\TRACE
if LEVEL >= 100 then return;	\Don't allow rays to reflect forever
LEVEL:= LEVEL +1;		\Next level of recursion down

DISTMIN:= BIG;			\Infinitely far away (more or less)
for OBJ:= 0, OBJECTS-1 do	\For all objects see if ray intersects them
	begin
	DIST:= INTSPHERE(RP, RV, SPCENTER(OBJ), SPRADIUS(OBJ));
	if DIST > SMALL then	\(Prevent infinite reflections due to rounding errors)
	    if DIST < DISTMIN then	\Ray intersects -- save closest point
		begin
		DISTMIN:= DIST;
		COLOR:= SPCOLOR(OBJ);	\Save color of closest point
		OBJMIN:= OBJ;
		end;
	end;

if DISTMIN = BIG then
	begin		\Ray didn't hit sphere. See if it hits the bottom plane.
	DIST:= INTPLANE(RP, RV, 3500.0); \3500 = Y distance to bottom plane
	if DIST > SMALL then	\(Prevent infinite reflections due to rounding errors)
	    if DIST < DISTMIN then \Ray intersects -- save closest point
		begin
		DISTMIN:= DIST;
		CHECKER(DIST);	\Get color of this closest point
		end;
	end;

if COLOR = MIRROR then	\Reflect ray and trace it
	begin		\(This does not check for reflections off planes yet)
	RP2:= RLRES(3);   RV2:= RLRES(3);   NORM:= RLRES(3);

	\RP2 = RP + RV *DISTMIN		Point where ray intersects sphere
	RP2(0):= RP(0) + RV(0) *DISTMIN;
	RP2(1):= RP(1) + RV(1) *DISTMIN;
	RP2(2):= RP(2) + RV(2) *DISTMIN;

	\NORM:= (RP2 - C) /R	Normal to sphere at point of intersection
	NORM(0):= (RP2(0) - SPCENTER(OBJMIN,0)) /SPRADIUS(OBJMIN);
	NORM(1):= (RP2(1) - SPCENTER(OBJMIN,1)) /SPRADIUS(OBJMIN);
	NORM(2):= (RP2(2) - SPCENTER(OBJMIN,2)) /SPRADIUS(OBJMIN);

	REFLECT(RV, NORM, RV2);		\(Incident, Normal, Reflected)

	COLOR:= GRAY;			\Background color for reflections
	TRACE(RP2, RV2);		\Trace reflected ray
	end;
LEVEL:= LEVEL -1;			\Up a level of recursion
end;	\TRACE



begin	\TRACERAYS
RAYV:= RLRES(3);
LEVEL:= 0;
FIRSTTIME:= true;
SIZE:= SIZE0;

loop	begin
	YF:= false;
	Y:= 0;
	while Y < TVH do
		begin
		if KEYHIT then DOKEY;			\Check for key commands
		if REDRAW then return;
		XF:= false;
		X:= 0;
		while X < TVW do
			begin
			if XF ! YF ! FIRSTTIME then	\(Resolution control)
				begin	\Project ray from viewer to pixel X,Y
				RAYV(0):= (FLOAT(X) - EYE(0) + XOFFSET) /GAIN;
				RAYV(1):= (FLOAT(Y) - EYE(1) + YOFFSET) /GAIN;
				RAYV(2):= \0.0\ - EYE(2);
				UNITIZE(RAYV);		\Make it a unit vector
							\(Note RAYV is changed)
				COLOR:= BLACK;		\Background color
				TRACE(EYE, RAYV);	\Trace ray and get COLOR
				BLOCK(X, Y, COLOR, SIZE); \Plot COLOR
				end;
			X:= X +SIZE;			\Skip SIZE pixels
			XF:= ~XF;
			end;
		Y:= Y +SIZE;				\Skip SIZE pixels
		YF:= ~YF;
		end;
	SIZE:= SIZE >>1;				\Chop block size in half
	if SIZE = 0 then quit;
	FIRSTTIME:= false;
	end;
end;	\TRACERAYS

\----------------------------------------------------------------------

proc	INITIALIZE;
begin
CPUREG:= GETREG;	\Get registers so we can call BIOS routines
MODE0:= GETVID;		\Save current video mode so it can be restored upon exit
SETVID($12);		\Set VGA mode (also see def TVW, TVH)

\Make some objects (spheres):
OBJECTS:= 3;				\Number of objects (spheres)
SPCOLOR:= [RED, MIRROR, MIRROR];	\Color of each sphere
SPRADIUS:= [200.0, 300.0, 600.0];	\Radii
SPCENTER:=[[400.0, 700.0, 1200.0],	\Centers (X,Y,Z)
	   [-100.0, 400.0, 1640.0],
	   [700.0, 170.0, 2000.0]];

EYE:= [320.0, 240.0, -1500.0];		\Coordinates of viewer (X,Y,Z)

GAIN:= 1.00;   XOFFSET:= -30.0;   YOFFSET:= 90.0;
end;	\INITIALIZE

\----------------------------------------------------------------------

begin	\MAIN
INITIALIZE;

loop	begin
	loop	begin
		REDRAW:= false;		\This flag is set to true by DOKEY
		TRACERAYS;
		if ~REDRAW then quit;
		end;
	CHOUT(0, BEL);	\Done at last!
	DOKEY;		\Wait for key press to exit or redraw
	end;
end;	\MAIN
